\subsubsection{Configuration of the GUI}\label{subsubsec:configuration}
Functions and constants used to find and load the glade file and manage the 
most recently used lists.
\begin{code}
module Emulator.Configuration.Configuration(
                                             getFromGlade,
                                             updateMRUList,
                                             baseTitle,
                                             getMRUs) where

import Graphics.UI.Gtk
import Graphics.UI.Gtk.Glade
import Graphics.UI.Gtk.Types as Gtypes


import Data.IORef
import Data.List as List
import Data.Maybe

import System.Directory (doesFileExist,getHomeDirectory,createDirectoryIfMissing)
import System.FilePath
import System.IO.Error
import System.IO

import Emulator.Data.GUI

import Paths_lqpl

baseTitle :: String 
baseTitle = "Quantum Stack Machine"


gladeFileName :: String 
gladeFileName = "qstack.glade"
subdir :: String 
subdir = "glade/"
homeconfdir :: String 
homeconfdir = "/.lqpl/"

mruconffile :: String
mruconffile = "MRU"


updateMRUList :: String -> IO()
updateMRUList newfile
    = do (_,files) <- getMRUs
         let fls = nub $ take 10 $ newfile:files
             (cdir',_) = splitFileName newfile 
         setMRUs cdir' fls

mru :: IO String
mru = do  hdir <- getHomeDirectory
          let  cnfdir = hdir ++ homeconfdir
               mruFile = cnfdir ++ mruconffile
          mruExists <- doesFileExist mruFile
          if mruExists then return mruFile else 
               do  createDirectoryIfMissing True cnfdir
                   hMRU <- openFile mruFile WriteMode
                   hPutStr hMRU ""
                   hClose hMRU
                   return mruFile



getMRUs :: IO (Maybe String, [String])
getMRUs =  do  fnm <- mru
               hMRU <- openFile fnm ReadMode
               contents <- hGetLines hMRU
               case  contents of
                     hd:rst  -> return (Just hd,rst)
                     []      -> return (Nothing,[])

setMRUs :: String -> [String] -> IO()
setMRUs cdir fls = do  fnm <- mru
                       hMRU <- openFile fnm WriteMode
                       hPutStr hMRU $ unlines $ cdir:fls
                       hClose hMRU


fileLoc :: IO FilePath
fileLoc = 
    catch (getDataFileName (subdir++gladeFileName))
              (\ _ -> return  gladeFileName)

retrieveGladeXML :: FilePath -> IO (GladeXML)
retrieveGladeXML fpath
    = do qsxml <- xmlNew fpath
         case qsxml of
	       Just qsXml -> return qsXml
	       Nothing ->
                   error $ "Unable to find glade file in "++fpath

                                  
getGladeXML :: IO (GladeXML)
getGladeXML =  fileLoc >>= retrieveGladeXML



getFromGlade :: IO (PrimeGUI, SubDisplayGUI,SubDisplayGUI,SubDisplayGUI,LQPLWindowGUI,PrefDialogGUI,CompilerPrefDialogGUI)
getFromGlade =
    do qsXml <- getGladeXML
       mainW <- xmlGetWidget qsXml castToWindow "qsMainWindow"
       scroller <- xmlGetWidget qsXml castToScrolledWindow "qswScrolling"
       qsArea <- xmlGetWidget qsXml castToDrawingArea "qswArea"
       progBar <-   xmlGetWidget qsXml castToProgressBar "executionProgress"  
       showTraceCB <-xmlGetWidget qsXml castToCheckButton "showTrace"
       
       [stepSpin, sdSpin, tdSpin] <- 
           mapM (xmlGetWidget qsXml castToSpinButton)
                    ["stepCountSpinButton","streamDepthSpin","treeDepthSpin"]

       cstackLabel <- xmlGetWidget qsXml castToLabel "classicalStackLabel"
       [trimButton, stepButton, goButton] <- 
            mapM (xmlGetWidget qsXml castToButton) 
                     ["trimStackButton","stepButton","goButton"]
       [simMenuItem, fileMenu, openMenuItem, openLQPLMI, rofMI,
        rsfMI, snapshotMenuItem, prefMenuItem, compilerPrefMenuItem,
        quitMenuItem, aboutMenuItem] <- 
              mapM (xmlGetWidget qsXml castToMenuItem)  
                       ["simulate1","fileMenu","open","openLQPL",
                        "recentObjectFiles","recentSourceFiles","snapshotMI",
                        "preferences_menu_item","compiler_pref_menu_item",
                        "quit","about"]
       
       vmDialog <- xmlGetWidget qsXml castToDialog "versionWarningDialog"
       cvText<- xmlGetWidget qsXml castToLabel "compilerVersionLabel"
       svText <- xmlGetWidget qsXml castToLabel "emulatorVersionLabel"
       widgetHideAll vmDialog
       simDialog <- xmlGetWidget qsXml castToDialog "simulateDialog"
       qsAbout  <- xmlGetWidget qsXml  castToAboutDialog "aboutdialog"
       widgetHideAll qsAbout
       rvText <- xmlGetWidget qsXml castToLabel "randomValText"
       resTable <- xmlGetWidget qsXml castToTable "resultsTable"
       widgetHideAll simDialog

       let mainGui = PrimeGUI mainW scroller qsArea progBar showTraceCB stepSpin sdSpin tdSpin 
                     cstackLabel trimButton stepButton goButton simMenuItem fileMenu openMenuItem
                     openLQPLMI rofMI rsfMI snapshotMenuItem prefMenuItem compilerPrefMenuItem aboutMenuItem 
                     quitMenuItem  qsAbout 
                     vmDialog cvText svText
                     simDialog rvText resTable

       mmw <- xmlGetWidget qsXml  castToWindow "MMapWindow"
       mmTextView <- xmlGetWidget qsXml castToTextView "mapElementsTextView"
       widgetHideAll mmw
       showHideMM <- xmlGetWidget qsXml castToMenuItem "hide_or_show_mmap"
       let mmGui = SubDisplayGUI mmw (Left mmTextView) showHideMM

       dumpWindow <- xmlGetWidget qsXml  castToWindow "DumpWindow"
       dumpTextView <- xmlGetWidget qsXml castToTextView "dumpTextView"
       widgetHideAll dumpWindow
       showHideDumpMI <-xmlGetWidget qsXml castToMenuItem "hide_or_show_dump"
       let dumpGui = SubDisplayGUI dumpWindow (Left dumpTextView) showHideDumpMI 

       codeWindow <- xmlGetWidget qsXml  castToWindow "CodeWindow"
       instructionTabs <- xmlGetWidget qsXml castToNotebook "instructionLists"
       showHideCodeMI <-xmlGetWidget qsXml castToMenuItem "hide_or_show_code"
       let qpoGui = SubDisplayGUI codeWindow (Right instructionTabs) showHideCodeMI 

       lqplWindow <- xmlGetWidget qsXml  castToWindow "LqplWindow"
       lqplTabs <- xmlGetWidget qsXml castToNotebook "lqplNotebook"
       showHideLqplMI <-xmlGetWidget qsXml castToMenuItem "hide_or_show_LQPL"
       compileBtn  <-xmlGetWidget qsXml castToButton "compileLQPL"
       let lqplGui = LQPLWindowGUI (SubDisplayGUI lqplWindow (Right lqplTabs) showHideLqplMI) compileBtn 
       
       prefDialog <- xmlGetWidget qsXml castToDialog "configQSWindowDialog"
       fontButtonFB <-  xmlGetWidget qsXml castToFontButton "fontbutton"
       [pdshowTraceCB,useEpsilonCB] <- 
             mapM (xmlGetWidget qsXml castToCheckButton) 
                      ["showTraceCheckButton","useEpsilonTrimming"]
       [nsSB, msSB, nmXoffsetSB, nmYoffsetSB,  trXoffsetSB, trYoffsetSB, 
            nameSizeSB, brLblSizeSB, traceSizeSB, elisionSizeSB, 
            leafSizeSB, tracePrecSB, leafPrecSB, cdSB, debugLevelSB, 
            epsilonAmountSB, trimRateSB] <- 
              mapM (xmlGetWidget qsXml castToSpinButton)  
                ["dotsizeSpinButton", "minSeparationSpinButton", "nameXoffsetSpinButton",
                 "nameYoffsetSpinButton", "traceXoffsetSpinButton", "traceYoffsetSpinButton",
                 "nameFontSizeSpinButton", "branchLabelFontSizeSpinButton",
                 "traceFontSizeSpinButton", "elisionFontSizeSpinButton", "leafFontSizeSpinButton",
                 "traceShowDigitsSpinButton", "leafShowDigitsSpinButton", "callDepthSpinButton",
                 "debugLevelSpinButton", "epsilonAmountSpinButton", "trimRateSpinButton"]
       let prefDGui = PrefDialogGUI prefDialog fontButtonFB pdshowTraceCB useEpsilonCB 
                      nsSB msSB nmXoffsetSB nmYoffsetSB trXoffsetSB trYoffsetSB 
                      nameSizeSB brLblSizeSB traceSizeSB elisionSizeSB 
                      leafSizeSB tracePrecSB leafPrecSB cdSB debugLevelSB 
                      epsilonAmountSB trimRateSB

       compPrefDialog <- xmlGetWidget qsXml castToDialog "compileOptionsDialog"
       cpFileChooserBtn <-  xmlGetWidget qsXml castToFileChooserButton "includeDirectoryChooser"
       cpincpathtv <- xmlGetWidget qsXml castToTreeView "includeDirectories"
       [showSynCB,showIRCB,showLoadCodeCB] <- 
             mapM (xmlGetWidget qsXml castToCheckButton) 
                      ["showSyntax","showIR", "loadCode"]
       [cplogEverything, cplogErrors, cplogWarnings, cplogInfo, cplogDebug, cplogFull] <-
              mapM (xmlGetWidget qsXml castToRadioButton)  
                ["logEverything", "logError", "logWarning", "logInfo",
                 "logDebug", "logTrace"]
       let compPrefDGui = CompilerPrefDialogGUI compPrefDialog showSynCB showIRCB showLoadCodeCB
                          cplogEverything cplogErrors cplogWarnings cplogInfo
                          cplogDebug cplogFull cpFileChooserBtn cpincpathtv
       return (mainGui,mmGui, dumpGui, qpoGui, lqplGui, prefDGui, compPrefDGui)
        

hGetLines :: Handle -> IO [String]
hGetLines h
    = do  isEOF <-  hIsEOF h
          isClosed <- hIsClosed h
          if isEOF || isClosed 
            then do  hClose h
                     return []
            else do  ln <- hGetLine h
                     lines <- hGetLines h
                     return $ ln : lines
\end{code}
